import uuid
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Literal

from fastapi_mongo_base.schemas import BusinessOwnedEntitySchema
from pydantic import BaseModel, ConfigDict, field_serializer, model_validator


class Currency(str, Enum):
    none = "none"

    IRR = "IRR"
    IRT = "IRT"

    USD = "USD"
    EUR = "EUR"
    GBP = "GBP"

    USDT = "USDT"
    BTC = "BTC"
    ETH = "ETH"


class WalletType(str, Enum):
    user = "user"
    business = "business"
    app = "app"
    app_operational = "app_operational"
    app_income = "app_income"


class WalletSchema(BusinessOwnedEntitySchema):
    pass


class WalletDetailSchema(BusinessOwnedEntitySchema):
    balance: dict[str, Decimal] = {}
    wallet_type: WalletType = WalletType.user
    main_currency: Currency

    model_config = ConfigDict(allow_inf_nan=True)

    @field_serializer("balance")
    def serialize_balance(self, balance: dict[str, Decimal]) -> dict[str, Decimal]:
        return {k: (v if v.is_finite() else Decimal(0)) for k, v in balance.items()}

    @field_serializer("wallet_type")
    def serialize_wallet_type(self, wallet_type: WalletType) -> str:
        return wallet_type.value

    @field_serializer("main_currency")
    def serialize_main_currency(self, main_currency: Currency) -> str:
        return main_currency.value


class WalletCreateSchema(BaseModel):
    user_id: uuid.UUID
    meta_data: dict | None = None

    wallet_type: WalletType = WalletType.user
    main_currency: Currency = Currency.none

    @model_validator(mode="before")
    def validate_wallet_type(cls, values):
        if values.get("wallet_type") and not values.get("main_currency"):
            raise ValueError("main_currency is required for income wallet")
        return values


class WalletUpdateSchema(BaseModel):
    meta_data: dict | None = None


class WalletHoldSchema(BusinessOwnedEntitySchema):
    wallet_id: uuid.UUID
    currency: str
    amount: Decimal
    expires_at: datetime
    status: str
    description: str | None = None


class WalletHoldCreateSchema(BaseModel):
    amount: Decimal
    expires_at: datetime
    status: str = "active"
    meta_data: dict | None = None
    description: str | None = None


class WalletHoldUpdateSchema(BaseModel):
    expires_at: datetime | None = None
    status: str | None = None
    meta_data: dict | None = None
    description: str | None = None


class TransactionSchema(BusinessOwnedEntitySchema):
    proposal_id: uuid.UUID
    wallet_id: uuid.UUID
    amount: Decimal
    currency: str
    balance: Decimal
    description: str | None = None
    note: str | None = None

    model_config = ConfigDict(allow_inf_nan=True)


class TransactionNoteUpdateSchema(BaseModel):
    note: str


class Participant(BaseModel):
    wallet_id: uuid.UUID
    amount: Decimal


class ProposalSchema(BusinessOwnedEntitySchema):
    issuer_id: uuid.UUID
    amount: Decimal
    description: str | None = None
    note: str | None = None
    currency: str
    # status: str
    task_status: str
    participants: list[Participant]


class ProposalCreateSchema(BaseModel):
    amount: Decimal
    description: str | None = None
    note: str | None = None
    currency: str
    task_status: Literal["draft", "init"] = "draft"
    participants: list[Participant]
    meta_data: dict | None = None


class ProposalUpdateSchema(BaseModel):
    # status: str | None
    task_status: Literal["init"] | None = None
    description: str | None = None
    note: str | None = None
    meta_data: dict | None = None



import os

import singleton
from fastapi_mongo_base.schemas import PaginatedResponse
from usso.session import AsyncUssoSession, UssoSession


class App(UssoSession):
    def __init__(
        self,
        app_name: str = "saas",
        *,
        ufaas_base_url: str = os.getenv("UFAAS_URL"),
        usso_base_url: str | None = os.getenv("USSO_URL"),
        api_key: str | None = os.getenv("UFAAS_API_KEY"),
        usso_refresh_url: str | None = os.getenv("USSO_REFRESH_URL"),
        refresh_token: str | None = os.getenv("USSO_REFRESH_TOKEN"),
        client: UssoSession | None = None,
    ):
        super().__init__(
            usso_base_url=usso_base_url,
            api_key=api_key,
            usso_refresh_url=usso_refresh_url,
            refresh_token=refresh_token,
            client=client,
        )
        if not ufaas_base_url and client and hasattr(client, "ufaas_base_url"):
            ufaas_base_url = client.ufaas_base_url
        if ufaas_base_url.endswith("/"):
            ufaas_base_url = ufaas_base_url[:-1]
        if not ufaas_base_url.endswith("/api/v1/apps"):
            ufaas_base_url = f"{ufaas_base_url}/api/v1/apps"
        self.ufaas_base_url = ufaas_base_url
        self.app_name = app_name
        self.app_url = f"{ufaas_base_url}/{app_name}/"
        self.initiate_resources()

    def initiate_resources(self, **kwargs):
        pass


class AsyncApp(AsyncUssoSession):

    def __init__(
        self,
        app_name: str = "saas",
        *,
        ufaas_base_url: str = os.getenv("UFAAS_URL"),
        usso_base_url: str | None = os.getenv("USSO_URL"),
        api_key: str | None = os.getenv("UFAAS_API_KEY"),
        usso_refresh_url: str | None = os.getenv("USSO_REFRESH_URL"),
        refresh_token: str | None = os.getenv("USSO_REFRESH_TOKEN"),
        client: AsyncUssoSession | None = None,
    ):
        super().__init__(
            usso_base_url=usso_base_url,
            api_key=api_key,
            usso_refresh_url=usso_refresh_url,
            refresh_token=refresh_token,
            client=client,
        )
        if not ufaas_base_url and client and hasattr(client, "ufaas_base_url"):
            ufaas_base_url = client.ufaas_base_url
        if ufaas_base_url.endswith("/"):
            ufaas_base_url = ufaas_base_url[:-1]
        if not ufaas_base_url.endswith("/api/v1/apps"):
            ufaas_base_url = f"{ufaas_base_url}/api/v1/apps"
        self.ufaas_base_url = ufaas_base_url
        self.app_name = app_name
        self.app_url = f"{ufaas_base_url}/{app_name}/"
        self.initiate_resources()

    def initiate_resources(self, **kwargs):
        pass


class Resource(UssoSession, metaclass=singleton.Singleton):
    def __init__(
        self,
        app_name: str = "saas",
        resource_name: str = "usages",
        schema: type = dict,
        *,
        ufaas_base_url: str = os.getenv("UFAAS_URL"),
        usso_base_url: str | None = os.getenv("USSO_URL"),
        api_key: str | None = os.getenv("UFAAS_API_KEY"),
        usso_refresh_url: str | None = os.getenv("USSO_REFRESH_URL"),
        refresh_token: str | None = os.getenv("USSO_REFRESH_TOKEN"),
        client: UssoSession | None = None,
    ):
        super().__init__(
            usso_base_url=usso_base_url,
            api_key=api_key,
            usso_refresh_url=usso_refresh_url,
            refresh_token=refresh_token,
            client=client,
        )
        if not ufaas_base_url and client and hasattr(client, "ufaas_base_url"):
            ufaas_base_url = client.ufaas_base_url
        if ufaas_base_url.endswith("/"):
            ufaas_base_url = ufaas_base_url[:-1]
        if not ufaas_base_url.endswith("/api/v1/apps"):
            ufaas_base_url = f"{ufaas_base_url}/api/v1/apps"
        self.ufaas_base_url = ufaas_base_url
        self.app_name = app_name
        self.resource_name = resource_name
        self.app_url = f"{ufaas_base_url}/{app_name}/"
        self.resource_url = f"{self.app_url}{resource_name}/"
        self._config_schemas(schema)

    def _config_schemas(self, schema: type = dict, **kwargs):
        self.schema = schema
        if schema == dict:
            self.list_response_schema = dict
        else:
            self.list_response_schema = PaginatedResponse[schema]
        self.list_item_schema = schema
        self.retrieve_response_schema = schema
        self.create_response_schema = schema
        self.update_response_schema = schema
        self.delete_response_schema = schema

    def list_items(self, offset: int = 0, limit: int = 10, **kwargs):
        resp = self.get(
            self.resource_url, params={"offset": offset, "limit": limit, **kwargs}
        )
        resp.raise_for_status()
        return self.list_response_schema(**resp.json())

    def retrieve_item(self, uid: str, **kwargs):
        resp = self.get(f"{self.resource_url}/{uid}", params=kwargs)
        resp.raise_for_status()
        return self.retrieve_response_schema(**resp.json())

    def create_item(self, obj: dict, **kwargs):
        resp = self.post(self.resource_url, json=obj, params=kwargs)
        resp.raise_for_status()
        return self.create_response_schema(**resp.json())

    def update_item(self, uid: str, obj: dict, **kwargs):
        resp = self.patch(f"{self.resource_url}/{uid}", json=obj, params=kwargs)
        resp.raise_for_status()
        return self.update_response_schema(**resp.json())

    def delete_item(self, uid: str, **kwargs):
        resp = self.delete(f"{self.resource_url}/{uid}", params=kwargs)
        resp.raise_for_status()
        return self.delete_response_schema(**resp.json())


class AsyncResource(AsyncUssoSession, metaclass=singleton.Singleton):

    def __init__(
        self,
        app_name: str = "saas",
        resource_name: str = "usages",
        schema: type = dict,
        *,
        ufaas_base_url: str = os.getenv("UFAAS_URL"),
        usso_base_url: str | None = os.getenv("USSO_URL"),
        api_key: str | None = os.getenv("UFAAS_API_KEY"),
        usso_refresh_url: str | None = os.getenv("USSO_REFRESH_URL"),
        refresh_token: str | None = os.getenv("USSO_REFRESH_TOKEN"),
        client: AsyncUssoSession | None = None,
    ):
        super().__init__(
            usso_base_url=usso_base_url,
            api_key=api_key,
            usso_refresh_url=usso_refresh_url,
            refresh_token=refresh_token,
            client=client,
        )
        if not ufaas_base_url and client and hasattr(client, "ufaas_base_url"):
            ufaas_base_url = client.ufaas_base_url
        if ufaas_base_url.endswith("/"):
            ufaas_base_url = ufaas_base_url[:-1]
        if not ufaas_base_url.endswith("/api/v1/apps"):
            ufaas_base_url = f"{ufaas_base_url}/api/v1/apps"
        self.ufaas_base_url = ufaas_base_url
        self.app_name = app_name
        self.resource_name = resource_name
        self.app_url = f"{ufaas_base_url}/{app_name}/"
        self.resource_url = f"{self.app_url}{resource_name}/"
        self._config_schemas(schema)

    def _config_schemas(self, schema: type = dict, **kwargs):
        self.schema = schema
        if schema == dict:
            self.list_response_schema = dict
        else:
            self.list_response_schema = PaginatedResponse[schema]
        self.list_item_schema = schema
        self.retrieve_response_schema = schema
        self.create_response_schema = schema
        self.update_response_schema = schema
        self.delete_response_schema = schema

    async def list_items(self, offset: int = 0, limit: int = 10, **kwargs):
        resp = await self.get(
            self.resource_url, params={"offset": offset, "limit": limit, **kwargs}
        )
        resp.raise_for_status()
        return self.list_response_schema(**resp.json())

    async def retrieve_item(self, uid: str, **kwargs):
        resp = await self.get(f"{self.resource_url}/{uid}", params=kwargs)
        resp.raise_for_status()
        return self.retrieve_response_schema(**resp.json())

    async def create_item(self, obj: dict, **kwargs):
        resp = await self.post(self.resource_url, json=obj, params=kwargs)
        resp.raise_for_status()
        return self.create_response_schema(**resp.json())

    async def update_item(self, uid: str, obj: dict, **kwargs):
        resp = await self.patch(f"{self.resource_url}/{uid}", json=obj, params=kwargs)
        resp.raise_for_status()
        return self.update_response_schema(**resp.json())

    async def delete_item(self, uid: str, **kwargs):
        resp = await self.delete(f"{self.resource_url}/{uid}", params=kwargs)
        resp.raise_for_status()
        return self.delete_response_schema(**resp.json())



import uuid
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Literal

from fastapi_mongo_base.schemas import BusinessOwnedEntitySchema
from pydantic import BaseModel, ConfigDict, Field


class Bundle(BaseModel):
    asset: str
    quota: Decimal
    unit: str | None = None

    model_config = ConfigDict(allow_inf_nan=True)


class AcquisitionType(str, Enum):
    trial = "trial"
    # credit = "credit"
    purchased = "purchased"
    gifted = "gifted"
    # deferred = "deferred"
    promotion = "promotion"
    # subscription = "subscription"
    # on_demand = "on_demand"
    borrowed = "borrowed"
    # freemium = "freemium"
    postpaid = "postpaid"


class EnrollmentCreateSchema(BaseModel):
    user_id: uuid.UUID
    bundles: list[Bundle]

    price: Decimal = Decimal(0)
    invoice_id: str | None = None
    start_at: datetime = Field(default_factory=datetime.now)
    expire_at: datetime | None = None
    duration: int | None = Field(None, description="Duration in days")
    status: Literal["active", "inactive"] = "active"
    acquisition_type: AcquisitionType = AcquisitionType.purchased

    variant: str | None = None
    meta_data: dict | None = None

    due_date: datetime | None = None


class EnrollmentSchema(EnrollmentCreateSchema, BusinessOwnedEntitySchema):
    paid_at: datetime | None = None
    leftover_bundles: list[Bundle] = []


class QuotaSchema(BaseModel):
    asset: str
    quota: Decimal
    unit: str | None = None
    variant: str | None = None
    _quota: Decimal | None = None

    model_config = ConfigDict(allow_inf_nan=True)


class UsageCreateSchema(BaseModel):
    user_id: uuid.UUID
    enrollment_id: uuid.UUID | None = None
    asset: str
    amount: Decimal = Decimal(1)
    variant: str | None = None
    meta_data: dict | None = None


class UsageConsumption(BaseModel):
    enrollment_id: uuid.UUID
    amount: Decimal
    leftover_bundles: list[Bundle] = []


class UsageSchema(BusinessOwnedEntitySchema):
    # enrollment_id: uuid.UUID
    # asset: str
    # amount: Decimal

    consumptions: list[UsageConsumption]
    asset: str
    amount: Decimal
    variant: str | None = None



from fastapi import Request
from fastapi.responses import JSONResponse

from ..exceptions import UFaaSException


async def ufaas_exception_handler(request: Request, exc: UFaaSException):
    return JSONResponse(
        status_code=exc.status_code,
        content={"message": exc.message, "error": exc.error},
    )


EXCEPTION_HANDLERS = {UFaaSException: ufaas_exception_handler}



error_messages = {}


class UFaaSException(Exception):
    def __init__(self, status_code: int, error: str, message: str = None):
        self.status_code = status_code
        self.error = error
        self.message = message
        if message is None:
            self.message = error_messages[error]
        super().__init__(message)


class InsufficientFunds(UFaaSException):
    def __init__(self, message: str = None):
        super().__init__(status_code=402, error="insufficient_funds", message=message)


class InvalidRequest(UFaaSException):
    def __init__(self, message: str = None):
        super().__init__(status_code=400, error="invalid_request", message=message)


class Unauthorized(UFaaSException):
    def __init__(self, message: str = None):
        super().__init__(status_code=401, error="unauthorized", message=message)


class Forbidden(UFaaSException):
    def __init__(self, message: str = None):
        super().__init__(status_code=403, error="forbidden", message=message)


class NotFound(UFaaSException):
    def __init__(self, message: str = None):
        super().__init__(status_code=404, error="not_found", message=message)



import os
from urllib.parse import urlparse
import singleton
from usso.session import AsyncUssoSession, UssoSession

from .apps.saas import AsyncSaaS, SaaS


class UFaaS(UssoSession):

    def __init__(
        self,
        *,
        ufaas_base_url: str = os.getenv("UFAAS_URL"),
        usso_base_url: str | None = os.getenv("USSO_URL"),
        api_key: str | None = os.getenv("UFAAS_API_KEY"),
        usso_refresh_url: str | None = os.getenv("USSO_REFRESH_URL"),
        refresh_token: str | None = os.getenv("USSO_REFRESH_TOKEN"),
        client: UssoSession | None = None,
    ):
        if usso_base_url is None:
            # calculate sso_url using ufiles_url
            # for example: media.pixiee.io/v1/f -> sso.pixiee.io
            # for example: media.ufaas.io/v1/f -> sso.ufaas.io
            # for example: media.pixy.ir/api/v1/f -> sso.pixy.ir
            # for example: storage.pixy.ir/api/v1/f -> sso.pixy.ir
            parsed_url = urlparse(ufaas_base_url)
            netloc = parsed_url.netloc
            netloc_parts = netloc.split(".")
            if len(netloc_parts) > 2:
                netloc_parts[0] = "sso"
            else:
                netloc_parts = ["sso", netloc]
            netloc = ".".join(netloc_parts)
            usso_base_url = f"https://{netloc}"

        super().__init__(
            usso_base_url=usso_base_url,
            api_key=api_key,
            usso_refresh_url=usso_refresh_url,
            refresh_token=refresh_token,
            client=client,
        )
        if not ufaas_base_url and client and hasattr(client, "ufaas_base_url"):
            ufaas_base_url = client.ufaas_base_url
        assert ufaas_base_url, "UFAAS_URL is required"
        if ufaas_base_url.endswith("/"):
            ufaas_base_url = ufaas_base_url[:-1]
        self.ufaas_base_url = ufaas_base_url
        self.headers.update({"accept-encoding": "identity"})

        self.initiate_apps()

    def initiate_apps(self):
        self.saas = SaaS(client=self)


class AsyncUFaaS(AsyncUssoSession):

    def __init__(
        self,
        *,
        ufaas_base_url: str = os.getenv("UFAAS_URL"),
        usso_base_url: str | None = os.getenv("USSO_URL"),
        api_key: str | None = os.getenv("UFILES_API_KEY"),
        usso_refresh_url: str | None = os.getenv("USSO_REFRESH_URL"),
        refresh_token: str | None = os.getenv("USSO_REFRESH_TOKEN"),
        client: AsyncUssoSession | None = None,
    ):
        if usso_base_url is None:
            # calculate sso_url using ufiles_url
            # for example: media.pixiee.io/v1/f -> sso.pixiee.io
            # for example: media.ufaas.io/v1/f -> sso.ufaas.io
            # for example: media.pixy.ir/api/v1/f -> sso.pixy.ir
            # for example: storage.pixy.ir/api/v1/f -> sso.pixy.ir
            parsed_url = urlparse(ufaas_base_url)
            netloc = parsed_url.netloc
            netloc_parts = netloc.split(".")
            if len(netloc_parts) > 2:
                netloc_parts[0] = "sso"
            else:
                netloc_parts = ["sso", netloc]
            netloc = ".".join(netloc_parts)
            usso_base_url = f"https://{netloc}"
            
        super().__init__(
            usso_base_url=usso_base_url,
            api_key=api_key,
            usso_refresh_url=usso_refresh_url,
            refresh_token=refresh_token,
            client=client,
        )
        if not ufaas_base_url and client and hasattr(client, "ufaas_base_url"):
            ufaas_base_url = client.ufaas_base_url
        assert ufaas_base_url, "UFAAS_URL is required"
        if ufaas_base_url.endswith("/"):
            ufaas_base_url = ufaas_base_url[:-1]
        self.ufaas_base_url = ufaas_base_url
        self.headers.update({"accept-encoding": "identity"})
        self.initiate_apps()

    def initiate_apps(self):
        self.saas = AsyncSaaS(client=self)



import hashlib
from io import BytesIO


def calculate_file_hash(file: BytesIO) -> str:
    file_hash = hashlib.md5()
    file.seek(0)  # Ensure we start from the beginning of the file
    while chunk := file.read(8192):
        file_hash.update(chunk)
    file.seek(0)  # Reset file pointer to the beginning
    return file_hash.hexdigest()
